use serde_json::{Map, Number, Value};
use std::str::FromStr;

use crate::constants::{ArrayInnerType, ValueType, BOOLEAN_KEYS, KEY_VALUE_TYPES};

pub fn convert_value(value: &str, key: Option<&str>) -> Value {
    let parsed: Value = parse_number(value);

    if KEY_VALUE_TYPES.contains_key(key.unwrap_or_default()) {
        match KEY_VALUE_TYPES[key.unwrap_or_default()] {
            ValueType::Array(separator, inner_type) => {
                split_or_single_string(value, separator, &inner_type)
            }
            ValueType::ChunkMap(separator) => chunk_map(value, separator),
        }
    } else if BOOLEAN_KEYS.contains(&key.unwrap_or_default()) {
        number_to_bool(parsed)
    } else {
        parsed
    }
}

pub fn parse_number(input: &str) -> Value {
    if let Ok(int_val) = input.parse::<i64>() {
        Value::Number(Number::from(int_val))
    } else if let Ok(float_val) = input.parse::<f64>() {
        Number::from_f64(float_val)
            .map(Value::Number)
            .unwrap_or_else(|| Value::String(input.to_string()))
    } else {
        Value::String(input.to_string())
    }
}

pub fn number_to_bool(value: Value) -> Value {
    match value {
        Value::Number(num) => {
            if let Some(float_val) = num.as_f64() {
                Value::Bool(float_val.trunc() as i64 >= 1)
            } else {
                Value::Bool(false)
            }
        }
        _ => Value::Bool(false),
    }
}

pub fn chunk_map(value: &str, separator: char) -> Value {
    let parts: Vec<&str> = value.split(separator).collect();

    if parts.len() % 2 != 0 {
        Value::String(value.to_string())
    } else {
        let mut map: Map<String, Value> = Map::new();

        for chunk in parts.chunks(2) {
            if let (Ok(key), value) = (i64::from_str(chunk[0]), chunk[1].to_string()) {
                map.insert(key.to_string(), Value::String(value));
            }
        }

        Value::Object(map)
    }
}

pub fn split_or_single_string(input: &str, separator: char, inner_type: &ArrayInnerType) -> Value {
    let values: Vec<Value> = if input.contains(separator) {
        input
            .split(separator)
            .map(|s| match inner_type {
                // ArrayInnerType::String => Value::String(s.to_string()), // ! Uncomment if needed
                ArrayInnerType::Number => parse_number(s),
            })
            .collect()
    } else {
        vec![match inner_type {
            // ArrayInnerType::String => Value::String(input), // ! Uncomment if needed
            ArrayInnerType::Number => parse_number(input),
        }]
    };

    Value::Array(values)
}
