use super::super::constants::*;
use super::convert::{convert_value, number_to_bool};
use super::decrypt::decode_base64;
use quick_xml::{escape::unescape, events::Event, reader::Reader};
use serde_json::{Map, Value};

pub fn validate_xml(xml: &[u8]) -> Result<(), &str> {
    const XML_HEADER: &[u8] = b"<?xml version=\"1.0\"?>";

    if xml.starts_with(XML_HEADER) {
        Ok(())
    } else {
        Err("Invalid/missing XML header")
    }
}

pub fn parse_save(data: &[u8]) -> Value {
    let mut reader: Reader<&[u8]> = Reader::from_reader(data);
    reader.config_mut().trim_text(true);

    let mut buf: Vec<u8> = Vec::with_capacity(data.len());

    parse_dict(None, None, &mut reader, &mut buf)
}

fn parse_dict(
    parent_parent_key: Option<&str>,
    parent_key: Option<&str>,
    reader: &mut Reader<&[u8]>,
    buf: &mut Vec<u8>,
) -> Value {
    let mut map: Map<String, Value> = Map::new();
    let mut current_key: Option<String> = None;
    let mut current_kcek: Option<&str> = None;

    loop {
        match reader.read_event_into(buf) {
            Err(e) => panic!("Error at position {}: {:?}", reader.error_position(), e),
            Ok(Event::Eof) => break,
            Ok(Event::Start(ref e)) => match e.name().as_ref() {
                b"k" => {
                    let key: &str = &read_text(reader, buf, &mut None, &mut current_kcek, None);

                    current_key = match key {
                        key if GAMESAVE_KEYS.contains_key(key) => {
                            Some(GAMESAVE_KEYS[key].to_string())
                        }
                        key if GAME_EVENTS.contains_key(key) => Some(GAME_EVENTS[key].to_string()),
                        key if GAME_VARIABLES.contains_key(key) => {
                            Some(GAME_VARIABLES[key].to_string())
                        }
                        key if parent_key.unwrap_or_default() == "stats" => {
                            let stat_key: &str = if STAT_KEYS.contains_key(key) {
                                STAT_KEYS[key]
                            } else {
                                key
                            };

                            if stat_key.starts_with("unique_") {
                                let parts: Vec<&str> = stat_key.split("_").collect::<Vec<&str>>();

                                match parts.len() {
                                    3 => Some(format!(
                                        "secret_coin_{}_{}",
                                        OFFICIAL_LEVEL_NAMES[parts[1]], parts[2]
                                    )),
                                    _ => Some(stat_key.to_string()),
                                }
                            } else {
                                Some(stat_key.to_string())
                            }
                        }
                        key if parent_key.unwrap_or_default() == "game_values" => {
                            let game_value_key: &str = if GAME_VARIABLES.contains_key(key) {
                                GAME_VARIABLES[key]
                            } else {
                                key
                            };

                            if ["c0_", "c1_"].iter().any(|x| game_value_key.starts_with(x)) {
                                let parts: Vec<&str> =
                                    game_value_key.split("_").collect::<Vec<&str>>();

                                match parts.len() {
                                    2 => Some(format!(
                                        "colour_{}_{}",
                                        match game_value_key {
                                            key if key.starts_with("c0_") => "1",
                                            key if key.starts_with("c1_") => "2",
                                            _ => "unknown",
                                        },
                                        parts[1]
                                    )),
                                    _ => Some(game_value_key.to_string()),
                                }
                            } else {
                                Some(game_value_key.to_string())
                            }
                        }
                        key if ["official_level_progress", "official_levels"]
                            .contains(&parent_key.unwrap_or_default()) =>
                        {
                            if OFFICIAL_LEVEL_NAMES.contains_key(key) {
                                Some(OFFICIAL_LEVEL_NAMES[key].to_string())
                            } else {
                                Some(key.to_string())
                            }
                        }
                        key if parent_key.unwrap_or_default() == "enabled_items" => {
                            let parts: Vec<&str> = key.split("_").collect::<Vec<&str>>();

                            match parts.len() {
                                2 => Some(format!("{}_{}", ICON_FORMS[parts[1]], parts[0])),
                                _ => Some(key.to_string()),
                            }
                        }
                        key if parent_key.unwrap_or_default() == "daily_rewards" => {
                            let parts: Vec<&str> = key.split("_").collect::<Vec<&str>>();

                            match parts.len() {
                                2 => Some(format!("{}_{}", CHEST_IDS[parts[0]], parts[1])),
                                _ => Some(key.to_string()),
                            }
                        }
                        key if ["amount", "local_levels", "local_lists"]
                            .contains(&parent_key.unwrap_or_default())
                            && key.starts_with("k_") =>
                        {
                            let replaced_key = key.replace("k_", "");
                            Some(replaced_key)
                        }
                        key if parent_key.unwrap_or_default() == "shop_purchases"
                            && SPECIAL_SHOP_PURCHASES.contains_key(key) =>
                        {
                            Some(SPECIAL_SHOP_PURCHASES[key].to_string())
                        }
                        key if parent_key.unwrap_or_default() == "wraith_rewards"
                            && WRAITH_KEYS.contains_key(key) =>
                        {
                            Some(WRAITH_KEYS[key].to_string())
                        }
                        key if current_kcek.is_some() => match *current_kcek.as_ref().unwrap() {
                            "level" => Some(LEVEL_KEYS[key].to_string()),
                            "song" => Some(SONG_KEYS[key].to_string()),
                            "quest" => Some(QUEST_KEYS[key].to_string()),
                            "reward" => Some(REWARD_KEYS[key].to_string()),
                            "reward_data" => Some(REWARD_DATA_KEYS[key].to_string()),
                            "smart_templates" => Some(SMART_TEMPLATE_KEYS[key].to_string()),
                            "lists" => Some(LIST_KEYS[key].to_string()),
                            _ => Some(key.to_string()),
                        },
                        _ => Some(key.to_string()),
                    }
                }
                b"s" | b"r" | b"i" => {
                    if let Some(key) = current_key.take() {
                        let value: String = read_text(
                            reader,
                            buf,
                            &mut Some(key.as_str()),
                            &mut current_kcek,
                            parent_parent_key,
                        );
                        map.insert(key.to_string(), convert_value(&value, Some(key.as_str())));
                    }
                }
                b"d" => {
                    if let Some(key) = current_key.take() {
                        let nested_dict: Value =
                            parse_dict(parent_key, Some(key.as_str()), reader, buf);

                        map.insert(
                            key.to_string(),
                            if BOOLEAN_PARENT_KEYS.contains(&key.as_str()) {
                                filter_true_keys(nested_dict)
                            } else {
                                nested_dict
                            },
                        );
                    }
                }
                _ => (),
            },
            Ok(Event::End(ref e)) if e.name().as_ref() == b"d" => break,
            _ => (),
        }

        buf.clear();
    }

    Value::Object(map)
}

fn filter_true_keys(object: Value) -> Value {
    if let Value::Object(map) = object {
        let keys_with_true_values: Vec<Value> = map
            .into_iter()
            .filter_map(|(key, value)| {
                if number_to_bool(value) == Value::Bool(true) {
                    Some(Value::String(key))
                } else {
                    None
                }
            })
            .collect();
        Value::Array(keys_with_true_values)
    } else {
        Value::Null
    }
}

fn read_text(
    reader: &mut Reader<&[u8]>,
    buf: &mut Vec<u8>,
    key: &mut Option<&str>,
    kcek: &mut Option<&str>,
    parent_parent_key: Option<&str>,
) -> String {
    match reader.read_event_into(buf) {
        Ok(Event::Text(e)) => {
            let text: String = unescape(String::from_utf8_lossy(&e).to_string().as_str())
                .expect("Failed to unescape text")
                .into_owned();

            if let Some(key) = key {
                return match *key {
                    "type" => {
                        if KCEK_KEYS.contains_key(text.as_str()) {
                            let kcek_type: &str = KCEK_KEYS[text.as_str()];
                            *kcek = Some(kcek_type);
                            kcek_type.to_string()
                        } else if parent_parent_key.unwrap_or_default() == "quests" {
                            QUEST_TYPE[text.as_str()].to_string()
                        } else {
                            text
                        }
                    }
                    "item_unlock_value" => {
                        if ICON_FORMS.contains_key(text.as_str()) {
                            ICON_FORMS[text.as_str()].to_string()
                        } else {
                            text
                        }
                    }
                    "youtube_channel" => format!("{}{}", YOUTUBE_CHANNEL_BASE_URL, text),
                    "difficulty" => DIFFICULTY[text.as_str()].to_string(),
                    "level_type" => LEVEL_TYPE[text.as_str()].to_string(),
                    "level_length" => LEVEL_LENGTH[text.as_str()].to_string(),
                    "epic_rating" => EPIC_RATING[text.as_str()].to_string(),
                    "item_type" => ITEMS[text.as_str()].to_string(),
                    "description" => decode_base64(&text),
                    "level_data" => {
                        if TRUNCATE_LEVEL_DATA {
                            "truncated to minimize size".to_string()
                        } else {
                            text
                        }
                    }
                    _ => text,
                };
            }

            text
        }
        Err(e) => panic!("Error at position {}: {:?}", reader.error_position(), e),
        _ => panic!("Failed to read text"),
    }
}
